#################################################
#################################################
##          COPYRIGHT - DATA SOCIETY           ##
#################################################
#################################################

################################
##    TEXT MINING : Day 3.    ##
################################

## NOTE: To run code blocks (separated by #%%), select somewhere inside the block 
##       and press ctrl + enter for Windows or Linux, 
##       and command + enter for Macs

#==================================================================================
#%%

# Outline:
#	Using a Tagger and Tagged Corpora
#	Mapping Words to Properties Using Python Dictionaries
#	Automatic Tagging
#	N-Gram Tagging

#==================================================================================
#%%
# Slide 4: Using a tagger

'''
A part-of-speech tagger, or POS-tagger, processes a sequence of words, and
 attaches a part of speech tag to each word (don't forget to import nltk):
'''

import nltk

# Import word_tokenize from nltk
from nltk import word_tokenize

text = word_tokenize("And now for something completely different")
nltk.pos_tag(text)


# NLTK provides documentation for each tag, which can be queried using the tag, 
# e.g. nltk.help.upenn_tagset('RB'), or a regular expression, 
# e.g. nltk.help.upenn_tagset('NN.*'). 
# Some corpora have README files with tagset documentation, 
# see nltk.corpus.???.readme(), substituting in the name of the corpus.

#==================================================================================
#%%
# Slide 5: Using a tagger

# What if we use homonyms (same word with a different meaning)?	
text = word_tokenize("They refuse to permit us to obtain the refuse permit")
nltk.pos_tag(text)


#==================================================================================
#%%
# Slide 6: Using a tagger

# How can we identify words that are used in similar context?
# First, we need to create the word list. 
text = nltk.Text(word.lower() for word in nltk.corpus.brown.words())

text.similar('woman')
text.similar('bought')
text.similar('over')
text.similar('the')


#==================================================================================
#%%
# Slide 7: Using a tagger

# We can create one of these special tuples from the standard string 
# representation of a tagged token, using the function str2tuple():
tagged_token = nltk.tag.str2tuple('fly/NN')
tagged_token
tagged_token[0]
tagged_token[1]


#==================================================================================
#%%
# Slide 8: Creating tagged tokens
 	
sent = '''
The/AT grand/JJ jury/NN commented/VBD on/IN a/AT number/NN of/IN
other/AP topics/NNS ,/, AMONG/IN them/PPO the/AT Atlanta/NP and/CC
Fulton/NP-tl County/NN-tl purchasing/VBG departments/NNS which/WDT it/PPS
said/VBD ``/`` ARE/BER well/QL operated/VBN and/CC follow/VB generally/RB
accepted/VBN practices/NNS which/WDT inure/VB to/IN the/AT best/JJT
interest/NN of/IN both/ABX governments/NNS ''/'' ./.
'''

[nltk.tag.str2tuple(t) for t in sent.split()]

#==================================================================================
#%%
# Slide 9: Built-in taggers

# We can use the tagged_words() syntax to view the tags of the words.
nltk.corpus.brown.tagged_words()

# The universal tagset contains different terms for parts of speech. 
nltk.corpus.brown.tagged_words(tagset = 'universal')

# If a corpus has tagged text, you'll be able to use the tagged_words() method.
print(nltk.corpus.nps_chat.tagged_words())

# You can learn more about the universal tagset here: 
# https://github.com/slavpetrov/universal-pos-tags

#==================================================================================
#%%
# Slide 10: Tabulating tags

# Make sure to import the Brown corpus first.
from nltk.corpus import brown

# Now we'll pull all the words in the "news" category with a universal tag.
brown_news_tagged = brown.tagged_words(categories = 'news', tagset = 'universal')

# We'll use the FreqDist() syntax from nltk to calculate
# the frequency distribution of the tags.
tag_fd = nltk.FreqDist(tag for (word, tag) in brown_news_tagged)

# What are the most common parts of speech?
tag_fd.most_common()


#==================================================================================
#%%
'''
Exercise 1:
Plot the above frequency distribution using tag_fd.plot(cumulative=True). 
What percentage of words are tagged using the first five tags of the above list?
'''





#==================================================================================
#%%
# Slide 14: Syntactic patterns: nouns

# Let's create bigrams with the bigrams() syntax from NLTK.
word_tag_pairs = nltk.bigrams(brown_news_tagged)

# Now we will identify the words that appear before a noun in the bigrams
# we just created. 
noun_preceders = [a[1] for (a, b) in word_tag_pairs if b[1] == 'NOUN']

# We'll use the FreqDist() method on the word list we
# just created to see which words occur most often before 
fdist = nltk.FreqDist(noun_preceders)

# Let's pull out only the part of speech term
# (not the frequency).
[tag for (tag, _) in fdist.most_common()]


#==================================================================================
#%%
# Slide 15: Syntactic patterns: verbs

# Let's create bigrams with the bigrams() syntax from NLTK.
word_tag_fd = nltk.FreqDist(brown_news_tagged)

# Now we will identify the most common words that are tagged as verbs in the 
# frequency distribution.
top_verbs = [wt[0] for (wt, _) in word_tag_fd.most_common() if wt[1] == 'VERB']

# We can pull the top terms from the 'top_verbs variable.
top_verbs[:10]


#==================================================================================
#%%
# Slide 16: Conditional frequency distribution

# First, we'll set up the Conditional Frequency Distribution
cfd1 = nltk.ConditionalFreqDist(brown_news_tagged)

# Let's look at some words that have multiple meanings to see how they're most 
# commonly used.
cfd1['yield'].most_common()

cfd1['cut'].most_common()


#==================================================================================
#%%
# Slide 17: Conditional frequency distribution

'''
We can reverse the order of the pairs, so that the tags are the conditions, 
and the words are the events. Now we can see likely words for a given tag. 

'''

# First, we'll set up the Conditional Frequency Distribution
brown_news_tagged_1 = brown.tagged_words(categories = 'news')	


cfd2 = nltk.ConditionalFreqDist((tag, word) for (word, tag) 
									  in brown_news_tagged_1)

# We can pull the top ten terms that 
# are tagged as "verb past participle" 
list(cfd2['VBN'])[:10]

#==================================================================================
#%%
# Slide 18: Conditional frequency distribution

'''
To clarify the distinction between VBD (past tense) and VBN (past participle), 
let's find words which can be both VBD and VBN, and see some surrounding text:
'''

# First, we'll create the list where verbs are labeled as both VBD and VBN
[w for w in cfd1.conditions() if 'VBD' in cfd1[w] and 'VBN' in cfd1[w]]

# Let's shorten the variable to 'bnt' to make it easier to work with.
bnt = brown_news_tagged_1

# Now, we can look at 'kicked' labeled as VBD and 
# pull out the indexed words before and after it 
# to see the words and parts of speech around it.
idx1 = bnt.index(('kicked', 'VBD'))
bnt[idx1 - 4: idx1 + 1]

# Let's see if the words are different for VBN.
idx2 = bnt.index(('kicked', 'VBN'))
bnt[idx2 - 4: idx2 + 1]


#==================================================================================
#%%
'''
Exercise 2:

Import a text file (either your own or the example .txt file provided) using: 

mport os
# os.chdir('D:')

f = open('document.txt', 'rU')
raw = f.read()

Now find:
1) The most common tag/ part of speech
2) Look at bigrams to find the most common words and their part of speech that 
   occur before whatever the most common part of speech from #1.
3) Find the most common verbs and sort them by frequency, with the most frequent first.
'''

#==================================================================================
#%%
# Slide 22: Default dictionaries

'''
Slide ---

Remember dictionaries? We are going to dive back into using a dictionary - 
but a special kind, a defaultdict.

This dictionary is able to automatically create an entry for this new key and 
give it a default value, such as 0 or the empty list.
'''

# Let's import defaultdict.
from collections import defaultdict

# Set the defaultdict to accept integers as the parameter, and name it 'frequency'.
frequency = defaultdict(int)

# We can define the term 'colorless' to have the parameter '4'. We can check it
# by running it – what happens when we enter a new term that's not defined?
frequency['colorless'] = 4
frequency['colorless']
frequency['ideas']

# Set the defaultdict to set lists as
# the parameter, and name it 'pos'.
pos = defaultdict(list)
pos['sleep'] = ['NOUN', 'VERB']
pos['ideas']


#==================================================================================
#%%
# Slide 23: Default dictionaries

# We first define the function that tells us if the POS is unknown, 
# label it as a noun
def POS():
    return 'NOUN'
POS()

# Let's set our default dictionary to the function 
# that will return a noun. Now, if we put in a 
# non-existent entry, it's automatically added.
pos = defaultdict(POS) 
pos['colorless'] = 'ADJ'
pos['blog'] 

# We can look at the list of 
# items from the dictionary
list(pos.items())

#==================================================================================
#%%
# Slide 25: Default dictionaries
 	
# Let's read in "Alice in Wonderland" and save it as 'alice'.
alice = nltk.corpus.gutenberg.words('carroll-alice.txt')

# We'll calculate the Frequency Distribution of words, and then subset out the 
# 1000 most common words.
vocab = nltk.FreqDist(alice)
v1000 = [word for (word, _) in vocab.most_common(1000)]

# Let's check the first 20 words of the set we just created.
v1000[:20]

#==================================================================================
#%%
# Slide 26: Default dictionaries

# We need to define a new function to automatically return the special
# token that we can define – let's call it 'UNK'.
def default_POS():
    return 'UNK'

# Set the defaultdict to return our special token and label it 'mapping'. Then,
# we will map 'v' to itself to assign
mapping = defaultdict(default_POS)
for v in v1000:
    mapping[v] = v
mapping

# Now, we'll use the mapping syntax again, but this time for all of Alice
# – now it will assign 'UNK' to any word that is not in the top 1000 common words.
alice2 = [mapping[v] for v in alice]
alice2[:100]

#==================================================================================
#%%

'''
Exercise 3:

Access one of the Shakespeare txt files from gutenberg. You first should display
 all fileids from gutenberg, how many Shakespeare files are there?

Now take one of them and create a default dictionary that labels every unknown 
part of speech as UNK.
'''



#==================================================================================
#%%
# Slide 28: Default dictionaries


# First, set the default dictionary for integers, and import brown.
counts = defaultdict(int)
from nltk.corpus import brown

# If the tag hasn't been seen before, it will have a zero count by default. Each time 
# we encounter a tag, we increment its count using the += operator.
for (word, tag) in brown.tagged_words(categories = 'news', tagset = 'universal'):
    	 counts[tag] += 1

counts['NOUN']


# We want to sort by frequency and POS. Check some of the frequencies yourself 
# and see if they meet the sorted criteria by frequency and POS.

#==================================================================================
#%%
# Slide 29: Default dictionaries

'''
We use itemgetter in an important idiom for sorting a dictionary by its values. 
This allows us to show words in decreasing order of frequency.
'''

# First, import itemgetter.
from operator import itemgetter

# The first parameter of sorted() is the items to sort, a list of tuples 
# consisting of a POS tag and a frequency. 
# The second parameter specifies the sort key using a function itemgetter().
# The last parameter of sorted() specifies that the items should be returned 
# in reverse order, i.e. decreasing values of frequency.

sorted(counts.items(), key = itemgetter(1), reverse = True)

# Let's set pair equal to a tuple and look at how we can pull out a particular
# object from the tuple.
pair = ('NP', 8336)
pair[1]

itemgetter(1)(pair)

#==================================================================================
#%%
'''
Exercise 4:

Using the same Shakespeare text you pulled earlier, build a dictionary that:
	1. Counts the frequency of each part of speech within the text
	2. Sorts the dictionary (using itemgetter) by frequency of part of speech, 
        decreasing in values.
'''



#==================================================================================
#%%
# Slide 33: Default dictionaries

'''
Here is another instance of using defaultdict, where we index words according 
to their last two letters:
'''


# First, set 'last_letters' as a default dictionary for lists
last_letters = defaultdict(list)

# Set the wordlist to the NLTK corpus in English
words = nltk.corpus.words.words('en')

# Build the for loop where the key is equal to the last two letters, and then, we'll 
# append the words to the matching key (which is the last two letters)
for word in words:
    key = word[-2:]
    last_letters[key].append(word)

# Let's try out our function!
last_letters['ly'][:10]
last_letters['zy'][:10]


#==================================================================================
#%%
# Slide 34: Default dictionaries

# First, set 'last_letters' as a default dictionary for lists.
anagrams = defaultdict(list)

# Let's set a for loop, where the key is equal the sorted letters in a word, and
# the event is equal to the word itself – this way, for every word that has the 
# same letters, it will have the same key.
for word in words:
    key = ''.join(sorted(word)) # Here we're sorting the letters of a word 
                                # and setting it equal to the key
    anagrams[key].append(word)  # Now, we're setting those sorted letters as
                                # the key, and appending the word as the event.

# Let's test it out.
anagrams['aeilnrt']

# Note: this will only work if we put the letters in alphabetical order
#       because the key is the sorted letters.

#==================================================================================
#%%
'''
Exercise 5:

Using the 'words' dictionary, find all words that start with the prefix 'un'.
Store it in a list named unwanted. What is the length of the list?
'''




#==================================================================================
#%%
# Slide 36: Default taggers

'''
We will now explore various ways to automatically add part-of-speech tags to text. 
We will see that the tag of a word depends on the word and its context within a sentence. 
For this reason, we will be working with data at the level 
of (tagged) sentences rather than words. We'll begin by loading the data we will be using.
'''

# First, set 'last_letters' as a default dictionary for lists.
from nltk.corpus import brown

# We'll set up the regular sentences and the tagged sentences to compare them. 
brown_tagged_sents = brown.tagged_sents(categories = 'news')
brown_sents = brown.sents(categories = 'news')


'''
Slide --- 

The Default Tagger
The simplest possible tagger assigns the same tag to each token. 
This may seem to be a rather banal step, 
but it establishes an important baseline for tagger performance. 
In order to get the best result, we tag each 
word with the most likely tag. Let's find out which tag is most 
likely (now using the unsimplified tagset):
'''

# Now we will see which tag is the most common in the text.
tags = [tag for (word, tag) in brown.tagged_words(categories = 'news')]
nltk.FreqDist(tags).max()

#==================================================================================
#%%
# Slide 37: Default taggers

# Now we can create a tagger that will tag all words as 'NN'.
raw = 'I do not like green eggs and ham, I do not like them Sam I am!'

# We'll tokenize the words in the sentence and then use the default tagger to 'NN'.
tokens = word_tokenize(raw)
default_tagger = nltk.DefaultTagger('NN')

# Let's check to make sure it worked.
default_tagger.tag(tokens)


# Now we can evaluate the method
default_tagger.evaluate(brown_tagged_sents)
# 0.13089484257215028

'''
Unsurprisingly, this method performs rather poorly. On a typical corpus, 
it will tag only about an eighth of the tokens correctly, as we see below:
'''

# Default taggers assign their tag to every single word, even words that have 
# never been encountered before. 
# As it happens, once we have processed several thousand words of English text, 
# most new words will be nouns. 
# As we will see, this means that default taggers can help to improve the 
# robustness of a language processing system. 

#==================================================================================
#%%
# Slide 39: Lookup taggers

# First we'll get the frequency distribution and the conditional
# frequency distribution for the Brown corpora categorized as 'news'.
fd = nltk.FreqDist(brown.words(categories = 'news'))
cfd = nltk.ConditionalFreqDist(brown.tagged_words(categories = 'news'))

# What does cfd look like? 
cfd

# Now we will store the most common words from the Brown corpus.
most_freq_words = fd.most_common(100)

#==================================================================================
#%%
# Slide 40: Lookup taggers

# We'll set 'likely_tags' equal to the max count of tags for the words that
# are in the 'most_freq_words' data - remember, since some of the words have
# multiple parts of speech, we only want to take the most frequent one.
likely_tags = dict((word, cfd[word].max()) for (word, _) in most_freq_words)

# Now, we can use the UnigramTagger() method and set it as the baseline tagger.
# We can see how much more accurate we are if we tag the sentences with
# the lookup tagger.
baseline_tagger = nltk.UnigramTagger(model = likely_tags)
baseline_tagger.evaluate(brown_tagged_sents)
# 0.45578495136941344

#==================================================================================
#%%
# Slide 41: Lookup taggers

# Let's test this out on untagged input text:
sent = brown.sents(categories = 'news')[3]
baseline_tagger.tag(sent)


#==================================================================================
#%%
# Slide 42: Lookup taggers
	
baseline_tagger = nltk.UnigramTagger(model=likely_tags,
                                     backoff=nltk.DefaultTagger('NN'))


#==================================================================================
#%%
# Slide 43: Lookup taggers

'''
Let's put all this together and write a program to create and evaluate lookup 
taggers having a range of sizes
'''

def performance(cfd, wordlist):
    lt = dict((word, cfd[word].max()) for word in wordlist)
    baseline_tagger = nltk.UnigramTagger(model=lt, backoff=nltk.DefaultTagger('NN'))
    return baseline_tagger.evaluate(brown.tagged_sents(categories = 'news'))

def display():
    import pylab
    word_freqs = nltk.FreqDist(brown.words(categories = 'news')).most_common()
    words_by_freq = [w for (w, _) in word_freqs]
    cfd = nltk.ConditionalFreqDist(brown.tagged_words(categories = 'news'))
    sizes = 2 ** pylab.arange(15)
    perfs = [performance(cfd, words_by_freq[:size]) for size in sizes]
    pylab.plot(sizes, perfs, '-bo')
    pylab.title('Lookup Tagger Performance with Varying Model Size')
    pylab.xlabel('Model Size')
    pylab.ylabel('Performance')
    pylab.show()

#==================================================================================
#%%
# Slide 44: Lookup taggers

display()

#==================================================================================
#%%
'''
Exercise 6:

Explain what the functions above are doing, in plain English. 
Write each line of each function, but in pseudo code.
'''




#==================================================================================
#%%
# Slide 46: Unigram taggers


# Let's import brown and set up the tagged sentences and regular sentences
# word lists. Then, we'll train the unigram tagger, use it to tag a sentence,
# and evaluate it.
from nltk.corpus import brown

brown_tagged_sents = brown.tagged_sents(categories = 'news')
brown_sents = brown.sents(categories = 'news')

# How does the unigram tagger perform on the sentences of 2007?
unigram_tagger = nltk.UnigramTagger(brown_tagged_sents)
unigram_tagger.tag(brown_sents[2007])

unigram_tagger.evaluate(brown_tagged_sents)
# 0.9349006503968017

# We train a UnigramTagger by specifying tagged sentence data as a parameter 
# when we initialize the tagger. 
# The training process involves inspecting the tag of each word and storing 
# the most likely tag for any word in a dictionary, stored inside the tagger.

#==================================================================================
#%%
# Slide 47: Separating training and testing data

'''
Slide --- 

Separating the Training and Testing Data

Now that we are training a tagger on some data, we must be careful not to test 
it on the same data, as we did in the above example. 
A tagger that simply memorized its training data and made no attempt to 
construct a general model would get a perfect score, 
but would also be useless for tagging new text. Instead, we should split 
the data, training on 90% and testing on the remaining 10%:
'''

size = int(len(brown_tagged_sents) * 0.9)
size
# 4160
train_sents = brown_tagged_sents[:size]
test_sents = brown_tagged_sents[size:]
unigram_tagger = nltk.UnigramTagger(train_sents)
unigram_tagger.evaluate(test_sents)
# 0.811721...

# Although the score is worse, we now have a better picture of the usefulness
# of this tagger, i.e. its performance on previously unseen text.




















